/**
 * Fanplayr Main Controller
 *
 * @module controllers/Fanplayr
 */

var ISML = require('dw/template/ISML');
var URLUtils = require('dw/web/URLUtils');
var CouponStatusCodes = require('dw/campaign/CouponStatusCodes');
var CouponLineItem = require('dw/order/CouponLineItem');
var OrderClass = require('dw/order/Order');

// needed by more than just Track now
var preferences = dw.system.Site.current.preferences.custom;
var forceHttps = preferences.fanplayrForceHttps || false;

/*
 * Called using /Fanplayr-Track
 */
var Track = function(){
	
	var trackDuringCheckout = preferences.fanplayrTrackDuringCheckout;
	
	// TODO: add secret security support
	//var fanplayrSecret = preferences.fanplayrSecret || null;
	
	var fanplayrEnabled = preferences.fanplayrIsEnable;
	if ( !fanplayrEnabled ) {
		return;
	}

    // -------------------------------------
    // Set up our tracking data
	var trackingData = new Object();
	
	// version #
	trackingData.mv = '20.3';
	
	// defaults shared across tracking types
	trackingData.products = new Array();
    trackingData.subTotal = 0;
    trackingData.total = 0;
    trackingData.discount = 0;
    trackingData.discountCode = '';
    trackingData.currency = '';
	
	trackingData.accountKey = preferences.fanplayrAccountKey;
	
	// -------------------------------------
	// Work out the page type
	var currentPipelineName = getCurrentPipelineName();
	
	// Set up the default page type
	var pageType = 'page';
	
	var checkoutStep = new Number(request.httpParameterMap.checkoutStep);
	var checkoutName = request.httpParameterMap.checkoutName;
	
	// page context can also be helpful in finding page type
	// types: home, search (category), product, cart, checkout, orderconfirmation
	var pageContextType = request.httpParameterMap.pageContextType ? request.httpParameterMap.pageContextType.value.toLowerCase() : null;
	// debug
	//trackingData.pageContextType = pageContextType;

	// Use the pipeline, pageContext and checkout info to set page type
	if ( currentPipelineName === 'Home-Show' || pageContextType === 'home' || pageContextType === 'storefront' ) {
		pageType = 'home';
	} else if ( currentPipelineName === 'Search-Show' || pageContextType === 'search' ) {
		pageType = 'cat';
	} else if ( currentPipelineName.indexOf('Product-') === 0 || pageContextType === 'product' ) {
        pageType = 'prod';
    } else if ( currentPipelineName === 'COSummary-ShowConfirmation' || currentPipelineName === 'COSummary-Submit' || pageContextType === 'orderconfirmation' || currentPipelineName === "Order-Confirm" || pageContextType === 'Order-Confirm' || (preferences.fanplayrOrderPipelineName && preferences.fanplayrOrderPipelineName === currentPipelineName) ) {
        pageType = 'order';
    } else if ( currentPipelineName === 'Cart-Show' || pageContextType === 'cart' ) {
        pageType = 'cart';
    }
	
	if ( pageType === 'cart' && checkoutStep > 0 || pageContextType === 'checkout' ) {
		pageType = 'checkout';
	}
	
	if ( pageType === 'checkout' && !trackDuringCheckout ) {
		return;
	}
	
	trackingData.pageType = pageType;

    // -------------------------------------
    // Get references to managers
	var orderMgr = dw.order.OrderMgr;
	var productMgr = dw.catalog.ProductMgr;
	var basketMgr = dw.order.BasketMgr;
	var taxMgr = dw.order.TaxMgr;

	// debug
	// trackingData.taxationPolicy = taxMgr.taxationPolicy == taxMgr.TAX_POLICY_GROSS ? 'GROSS' : 'NET';
	
	trackingData.connect = {
		endpoint: (forceHttps ? URLUtils.https('Fanplayr-Connect') : URLUtils.abs('Fanplayr-Connect')) + ''
	};
	
    // -------------------------------------
    if ( pageType !== 'order' ) {
        
        // set up defaults for user tracking
        trackingData.categoryId = '';
        trackingData.categoryName = '';
        trackingData.productId = '';
        trackingData.productName = '';
        trackingData.productSku = '';
        trackingData.productPrice = 0;
        trackingData.productUrl = '';
        trackingData.productImage = '';
        
        trackingData.lineItemCount = 0;
        trackingData.numItems = 0;
        
        // session is a global, but the order may use different code
        var currency = session.getCurrency()
        trackingData.currency = currency.currencyCode;
        
        // Collect baskset information
        var basket = basketMgr.currentBasket;
        
        if ( basket !== null ) {
            var dataAdded = addProductAndRelatedData(basket, trackingData, true);
            // TODO: Do we need to log an error here if it doesn't exist?
        }
        
        // If we are on a product page, get information about that product
        var productId;
        if (request.httpParameterMap && request.httpParameterMap.productId) {
            request.httpParameterMap.productId.value;
        }
        if ( productId !== null ) {
            var product = productMgr.getProduct(productId);
            
            trackingData.productSku = productId;
            
            if ( product !== null ) {
                if (preferences.fanplayrMasterProductTrackingEnable && product.isVariant()) {
                    trackingData.productId = product.getVariationModel().master.ID;
                }
                else {
                    trackingData.productId = productId;
                }
                trackingData.productName = product.name;
                
                // price
                var variant = product.isMaster() && product.variationModel.variants.size() > 0 && product.variationModel.variants[0];
                var priceModel = variant ? variant.getPriceModel() : product.getPriceModel();
                if ( priceModel ) {
                  var price = priceModel.price || priceModel.maxPrice || priceModel.minPrice;
                  trackingData.productPrice = price && price.value;
                }

                // category
                var category = product.primaryCategory;
                if ( !category && product.isVariant() && product.masterProduct ) {
                  category = product.masterProduct.primaryCategory;
                }
                if ( category ) {
                  trackingData.categoryId = category.ID;
                  trackingData.categoryName = category.displayName;
                }
                
                // url : dw.web.URL
                var productUrl = null;
                if ( product.primaryCategory && product.primaryCategory.ID != null ) {
                    productUrl = forceHttps ?
                    		URLUtils.https('Product-Show','pid', product.ID, 'cgid', product.primaryCategory.ID) :
                    		URLUtils.abs('Product-Show','pid', product.ID, 'cgid', product.primaryCategory.ID);
                } else {
                    productUrl = forceHttps ?
                    		URLUtils.https('Product-Show','pid', product.ID) :
                    		URLUtils.abs('Product-Show','pid', product.ID);
                }
                if (productUrl !== null) {
                    trackingData.productUrl = productUrl.toString();
                }
                
                // image : dw.web.URL
                var productImgUrl = null;
                //  : dw.content.MediaFile
                var productImage = product.getImage(preferences.fanplayrProductImageViewtype || 'large', 0);
                if ( productImage ) {
                    productImgUrl = productImage.getAbsURL();
                }
                if ( productImgUrl !== null ) {
                    trackingData.productImage = productImgUrl.toString();
                }
            }
        }
        
        // If we are on a category page, get information about that category
        var categoryId;
        if (request.httpParameterMap && request.httpParameterMap.categoryId) {
            categoryId = request.httpParameterMap.categoryId.value;
        }
        if ( categoryId ) {
            trackingData.categoryId = categoryId;
            trackingData.categoryName = categoryId;
        } else {
            // let's see if we can get it some other way?
            // but we will worry about this in the next release
        }
        
        // set URLs
        trackingData.applyToCartUrl = '';
        if ( preferences.fanplayrEnableApplyToCart ) {
            trackingData.applyToCartUrl = (forceHttps ? URLUtils.https('Fanplayr-Apply') : URLUtils.abs('Fanplayr-Apply')) + '?code=%c';
        }
        trackingData.sessionOfferUrl = '';
        if ( preferences.fanplayrEnableSessionCoupons ) {
            trackingData.sessionOfferUrl = (forceHttps ? URLUtils.https('Fanplayr-Session') : URLUtils.abs('Fanplayr-Session')) + '?code=%c';
            // TODO: add secret security feature
            /*if ( fanplayrSecret !== null ) {
                trackingData.sessionOfferUrl += '&auth=%a';
            }*/
        }
        
        // Render user tracking template
        ISML.renderTemplate('tags/FanplayrUserTracking', {
            FanplayrUserTrackingData: encodeURIComponent(JSON.stringify(trackingData).replace(/"/g, '\"')),
            FanplayrCustomLoader: preferences.fanplayrCustomLoader || ''
        });
        
    } else {
        // Order submitted page ('COSummary-Submit')
        
        // set up defaults for order tracking
        trackingData.shipping = 0;
        trackingData.tax = 0;
        trackingData.orderEmail = '';
        trackingData.firstName = '';
        trackingData.lastName = '';
        trackingData.customerEmail = '';
        trackingData.customerId = '';
        
        // We should have the order number sent through (from the PDict template variable)
        var orderNumber = request.httpParameterMap.orderNo ? request.httpParameterMap.orderNo.value : '';
        // : Order
        var order;
        
        // TODO: Maybe we should log an error if there was no order ?
        if ( orderNumber !== null ) {
            order = orderMgr.getOrder(orderNumber);
            // TODO: Maybe we should log an error if there was no order ?
            // we also make sure that it is a "NEW" order
            var passedNewOrderCheck = true;
            if ( preferences.fanplayrOnlyNewOrders && order !== null ) {
            	if ( order.getStatus().getValue() !== OrderClass.ORDER_STATUS_NEW ) {
            		passedNewOrderCheck = false;
            	}
            }
            if ( order !== null && passedNewOrderCheck ) {
                // it's best to track the order currency, rather than the current session
                // it may be different
                trackingData.currency = order.currencyCode;
                
                trackingData.orderNumber = orderNumber;
                
                // Order Number and Order ID are the same thing on this system
                trackingData.orderId = orderNumber;
                
                var dataAdded = addProductAndRelatedData(order, trackingData, false);
                
                // and now tax and shipping
                trackingData.tax = order.totalTax.available ? order.totalTax.getValue() : 0;
                trackingData.shipping = order.shippingTotalPrice.available ? order.shippingTotalPrice.getValue() : 0;
                
                // get customer info
                if ( customer !== null ) {
                    trackingData.customerId = customer.ID;
                    if ( customer.profile !== null ) {
                        trackingData.customerEmail = customer.profile.email ? customer.profile.email : '';
                        trackingData.firstName = customer.profile.firstName ? customer.profile.firstName : '';
                        trackingData.lastName = customer.profile.lastName ? customer.profile.lastName : '';
                    }
                }
                
                var orderEmail = order.getCustomerEmail();
                trackingData.orderEmail = orderEmail;
                
                // Render order tracking template
                ISML.renderTemplate('tags/FanplayrOrderTracking', {
                    FanplayrOrderTrackingData: encodeURIComponent(JSON.stringify(trackingData).replace(/"/g, '\"')),
                    FanplayrCustomLoader: preferences.fanplayrCustomLoader || ''
                });
            }
        }
    }

};

/**
 * Get the name of the current pipeline
 * Used for working out page type etc
 *
 * @returns name of the pipeline (hoping that ClickStream gives us the correct one)
 */
function getCurrentPipelineName() {
	// : List
	var clicks = session.clickStream.clicks;
	if (clicks.size() > 0) {
		// : ClickStreamEntry
		var click = clicks[clicks.size() - 1];
		return click.pipelineName;
	}
	return "";
}

/**
 * Get details including category ID, category name, URL and image URL
 * 
 * @param {Integer} productId
 * 
 * @returns an object like { image: STRING, url: STRING, catId: STRING, catName: STRING }
 */
function getProductExtras( productId ) {
	var data = {
		image: '',
		url: '',
		catId: '',
		catName: ''
	};
	
	var productMgr = dw.catalog.ProductMgr;
	var product = productMgr.getProduct(productId);
	
	if ( product !== null ) {
	    // category
	    var category = product.primaryCategory;
	    if ( !category && product.isVariant() && product.masterProduct ) {
	      category = product.masterProduct.primaryCategory;
	    }
	    if ( category ) {
	      data.catId = category.ID;
	      data.catName = category.displayName;
	    }
	    
	    // url : dw.web.URL
	    var productUrl = null;
	    if ( product.primaryCategory && product.primaryCategory.ID != null ) {
	        productUrl = forceHttps ?
	        		URLUtils.https('Product-Show','pid', product.ID, 'cgid', product.primaryCategory.ID) :
	        		URLUtils.abs('Product-Show','pid', product.ID, 'cgid', product.primaryCategory.ID);
	    } else {
	        productUrl = forceHttps ?
	        		URLUtils.https('Product-Show','pid', product.ID) :
	        		URLUtils.abs('Product-Show','pid', product.ID);
	    }
	    if (productUrl !== null) {
	        data.url = productUrl.toString();
	    }
	    
	    // image : dw.web.URL
	    var productImgUrl = null;
	    //  : dw.content.MediaFile
	    var productImage = product.getImage(preferences.fanplayrProductImageViewtype || 'large', 0); 
	    if ( productImage ) {
	        productImgUrl = productImage.getAbsURL();
	    }
	    if ( productImgUrl !== null ) {
	        data.image = productImgUrl.toString();
	    }
	}
	
	return data;
}

/**
 * Get discount codes applied
 * 
 * @param {LineItemCntr} lineItemContainer
 * 
 * @returns a comma separated list of coupon codes
 */
function getCurrentDiscountCodes( lineItemContainer ) {
    var coupons = new Array();
    if ( lineItemContainer === null || lineItemContainer.couponLineItems === null ) {
        return '';
    }
    for ( var i in lineItemContainer.couponLineItems ) {
        // : dw.order.CouponLineItem
    	var couponItem = lineItemContainer.couponLineItems[i];
        if ( couponItem.applied && couponItem.couponCode !== null ) {
            coupons.push(couponItem.couponCode);
        }
    }
    return coupons.join(',');
}

/**
 * Fills out an object with product and related information
 * 
 * @param {LineItemCntr} lineItemContainer
 * @param {Object} data
 * @param {Boolean} addLineDetails
 * 
 * @returns {Boolean} False if there were errors. Otherwise object has been filled.
 */
function addProductAndRelatedData( lineItemContainer, data, addLineDetails ) {
    
    if ( addLineDetails === undefined ) {
        addLineDetails = false;
    }
    
    if ( lineItemContainer === null ) {
        return false;
    }
    
    // get taxation policy
    var taxMgr = dw.order.TaxMgr;
    var useGross = taxMgr.taxationPolicy == taxMgr.TAX_POLICY_GROSS;
    
    data.subTotal = 0;
    data.products = new Array();
    
    if ( addLineDetails ) {
        data.lineItemCount = 0;
        data.numItems = 0;
    }
        
    // fill out product data
    for( var i in lineItemContainer.productLineItems ) {
        var prod = lineItemContainer.productLineItems[i];
        var extras = getProductExtras(prod.productID);
        
        var Id = prod.product.ID; 
        if (preferences.fanplayrMasterProductTrackingEnable && prod.product.isVariant()) {
        	Id = prod.product.getVariationModel().master.ID;
        }
        
        data.products.push({
        	id: Id,
//        	id: prod.product.master,
//        	catalogProduct: prod.catalogProduct,
//        	id: prod.productID,
            sku: prod.product.ID,
            name: prod.productName,
            qty: prod.quantity.value,
            price: (useGross ? prod.adjustedGrossPrice.value : prod.adjustedNetPrice.value) / prod.quantity.value,
            image: extras.image,
            url: extras.url,
            catId: extras.catId,
            catName: extras.catName
        });
        //data.subTotal += prod.adjustedNetPrice.value;
        if ( addLineDetails ) {
            data.lineItemCount++;
            data.numItems += prod.quantity.value;
        }
        
        // go through options and add those products
        for ( var j in prod.optionProductLineItems ) {
            var opt = prod.optionProductLineItems[j];
            var extras = getProductExtras(opt.productID);
            Id = opt.productID;
            if (preferences.fanplayrMasterProductTrackingEnable) {
                var optParent = opt.getParent();
                if (optParent && optParent.product && optParent.product.isVariant()) {
                    Id = optParent.product.getVariationModel().master.ID;
                }
            }
            data.products.push({
                id: Id,
                sku: opt.productID,
                name: opt.productName,
                qty: opt.quantity.value,
                price: (useGross ? opt.adjustedGrossPrice.value : opt.adjustedNetPrice.value) / opt.quantity.value,
                image: extras.image,
                url: extras.url,
                catId: extras.catId,
                catName: extras.catName
            });
            //data.subTotal += opt.adjustedNetPrice.value;
            if ( addLineDetails ) {
                data.lineItemCount++;
                data.numItems += opt.quantity.value;
            }
        }
    }
 
    // total, after discount
    data.subTotal = lineItemContainer.merchandizeTotalPrice.value;
    data.total = useGross ? lineItemContainer.adjustedMerchandizeTotalGrossPrice.value : lineItemContainer.adjustedMerchandizeTotalNetPrice.value;
    
    // let's get the discount here, we may be able to get it again later
    // using the discount line item, but let's do this just in case
    data.discount = (data.subTotal - data.total).toFixed(2);
    
    data.discountCode = getCurrentDiscountCodes(lineItemContainer);
    
    return true;
}

/*
 * Called using /Fanplayr-Apply/?code={coupon_code}
 * 
 * Tries to apply given coupon, and then redirects to the cart
 */
var ApplyCoupon = function() {
    
    var codeToApply = request.httpParameterMap.code.value && request.httpParameterMap.code.value.toLowerCase();
    var cartUrl = forceHttps ? URLUtils.https('Cart-Show') : URLUtils.abs('Cart-Show');
    
    var basketMgr = dw.order.BasketMgr;
    var basket = basketMgr.currentBasket;
    
    // let's check to see if this is a session coupon?
    // and if so if it can be applied
    var allowCoupon = true;
    
    var preferences = dw.system.Site.current.preferences.custom;
    var sessionCouponPrefix = preferences.fanplayrSessionCouponPrefix && preferences.fanplayrSessionCouponPrefix.toLowerCase() || null;
    
    if ( codeToApply && sessionCouponPrefix && codeToApply.indexOf(sessionCouponPrefix) === 0 ) {
        // get allowed session coupons from session
        var sessionCodesString = session && session.custom && session.custom.fanplayrSessionCodes || '[]';
        var sessionCodes = new Array();
        try {
            sessionCodes = JSON.parse(sessionCodesString);
        } catch (e) {} 
        // if the session coupon that the user has tried to apply
        // does not exist in the session them simply remove the code
        if ( sessionCodes.indexOf(codeToApply) === -1 ) {
            allowCoupon = false;
        }
    }
    
    // Is there a basket? Can we apply this coupon?
    if ( basket !== null && allowCoupon ) {
    
    	// : CouponLineItem
        var cli;
        // : Number
        var errorCode;
    
        try {
            var Transaction = require('dw/system/Transaction');
            Transaction.wrap(function() {
                cli = basket.createCouponLineItem(request.httpParameterMap.code.value, true);
            });
        } catch (e) {
            errorCode = e.errorCode;
        }
          
        if ( errorCode ) {
            // Error. Can error codes are here in CouponStatusCodes...
        } else {
            // Applied
        }
        
    } else {
        // No Baskset
    }
  
    response.redirect( cartUrl );
}

/*
 * Called using /Fanplayr-Session/?code={coupon_code}[&auth={auth}]
 * 
 * Tries to apply given coupon, and then redirects to the cart
 * If a secret has been set an md5'd salt+coupon_code+secret must be supplied
 * This makes sure that the request comes from Fanplayr and is specific to this coupon code
 * The salt is kept in the current session
 */
var SessionCoupon = function() {
    var codeToAdd = request.httpParameterMap.code.value && request.httpParameterMap.code.value.toLowerCase();
    var oldCodesString = session.custom.fanplayrSessionCodes || '[]';
    var oldCodes = new Array();
    try {
        oldCodes = JSON.parse(oldCodesString);
    } catch (e) {}
    if ( oldCodes.indexOf(codeToAdd) === -1 ) {
        oldCodes.push(codeToAdd);
    }

    session.custom.fanplayrSessionCodes = JSON.stringify(oldCodes);
    response.getWriter().println('{"error":false,"code":"'+codeToAdd+'"}');
    //response.getWriter().println(session.custom.fanplayrSessionCodes);
}

/*
 * Called using /Fanplayr-ServiceWorker/
 * 
 * Service Worker JS
 */
var ServiceWorker = function() {
	response.setHttpHeader("Content-Type", "text/javascript");
	response.getWriter().println("var fanplayrOptions = { version: 1, immediate: true };");
	response.getWriter().println("importScripts('https://static.fanplayr.com/client/sw.js?v=' + fanplayrOptions.version);");
}

/*
 * Called using /Fanplayr-Connect/
 * 
 * ITP compatability
 */
var Connect = function() {
	response.setHttpHeader("Content-Type", "application/json");
	
	var data = request.getHttpParameterMap().get('data').getStringValue();
	
	var cookieName = '_fphu';
	var Encoding = dw.crypto.Encoding;
	
	if (data) {
		// Store the data for 1 year.
		var cookie = new dw.web.Cookie(cookieName, Encoding.toURI(data));
		cookie.setMaxAge(31536000);
		cookie.setHttpOnly(true);
		cookie.setSecure(true);
		cookie.setPath('/');

		response.addHttpCookie(cookie);
		
		response.getWriter().println(data)
	} else {
		// Retrieve the data.
		var cookies = request.getHttpCookies();
		var cookieValue = cookies[cookieName] && cookies[cookieName].value
		response.getWriter().println(cookieValue ? Encoding.fromURI(cookieValue) : '');
	}
}

/*
 * Export functionality
 */
Track.public = true;
ApplyCoupon.public = true;
SessionCoupon.public = true;
ServiceWorker.public = true;
Connect.public = true;

module.exports = {
	Track: Track,
	Apply: ApplyCoupon,
	Session: SessionCoupon,
    ServiceWorker: ServiceWorker,
    Connect: Connect
};